from datetime import datetime
from typing import Type, List
from sqlalchemy import func, or_
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Character, CharacterStory, Activity, CharacterRelationship, CharacterTrait, Event, \
    CharacterEvent, Link, CharacterLink, Note, CharacterNote, Image, CharacterImage
from configparser import ConfigParser


class CharacterController(BaseController):
    """Character controller encapsulates characters management functionality

    Attributes
    ----------
    _instance : CharacterController
        The instance of the characters controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the characters controller
    _session : Session
        The database session

    Methods
    -------
    create_character(title: str, first_name: str, middle_name: str, last_name: str, honorific: str, nickname: str, \
                     gender: str, sex: str, ethnicity: str, race: str, tribe_or_clan: str, nationality: str, \
                     date_of_birth: str, date_of_death: str, description: str, mbti: str, enneagram: str, wounds: str)
        Create a new character
    update_character(character_id: int, title: str, first_name: str, middle_name: str, last_name: str, honorific: str, \
                     nickname: str, gender: str, sex: str, ethnicity: str, race: str, tribe_or_clan: str, \
                     nationality: str, date_of_birth: str, date_of_death: str, description: str, mbti: str, \
                     enneagram: str, wounds: str)
        Update a character
    delete_character(character_id: int)
        Delete a character
    get_character_by_id(character_id: int)
        Get a character by id
    get_character_count()
        Get character count associated with a user
    get_all_characters()
        Get all characters associated with a user
    get_all_characters_page(page: int, per_page: int)
        Get a single page of characters from the database associated with a user
    get_character_count_by_story_id(story_id: int)
        Get character count associated with a story
    get_characters_by_story_id(story_id: int)
        Get a list of characters by story id
    get_characters_page_by_story_id(story_id: int, page: int, per_page: int)
        Get a single page of characters by story id
    search_characters(search: str)
        Search for characters by title, first name, middle name, last name, nickname, and description belonging to a \
        specific user
    search_characters_by_story_id(story_id: int, search: str)
        Search for characters by title, first name, middle name, last name, nickname, and description belonging to a \
        specific story
    create_relationship(parent_id: int, related_id: int, relationship_type: str, description: str,
                        start_date: str, end_date: str)
        Create a new character relationship
    update_relationship(relationship_id: int, parent_id: int, related_id: int, relationship_type: str,
                        description: str, start_date: str, end_date: str)
        Update a character relationship
    change_relationship_position(relationship_id: int, position: int)
        Set the position of a character relationship
    delete_relationship(relationship_id: int)
        Delete a character relationship
    get_relationship_by_id(relationship_id: int)
        Get a character relationship by id
    get_relationships_by_character_id(parent_id: int)
        Get character relationships by character id, from that character's perspective
    get_relationships_page_by_character_id(parent_id: int, page: int, per_page: int)
        Get a single page of character relationships by character id, from that character's perspective
    create_trait(character_id: int, name: str, magnitude: int)
        Create a new character trait
    update_trait(trait_id: int, name: str, magnitude: int)
        Update a character trait
    change_trait_position(trait_id: int, position: int)
        Set the position of a character trait
    delete_trait(trait_id: int)
        Delete a character trait
    get_trait_by_id(trait_id: int)
        Get a character trait by id
    get_traits_by_character_id(character_id: int)
        Get character traits by character id
    get_traits_page_by_character_id(character_id: int, page: int, per_page: int)
        Get a single page of character traits by character id
    append_events_to_character(character_id: int, events: list)
        Append events to a character
    get_events_by_character_id(character_id: int)
        Get all events associated with a character
    get_events_page_by_character_id(character_id: int, page: int, per_page: int)
        Get a single page of events associated with a character from the database
    append_links_to_character(character_id: int, links: list)
        Append links to a character
    get_links_by_character_id(character_id: int)
        Get all links associated with a character
    get_links_page_by_character_id(character_id: int, page: int, per_page: int)
        Get a single page of links associated with a character from the database
    append_notes_to_character(character_id: int, notes: list)
        Append notes to a character
    get_notes_by_character_id(character_id: int)
        Get all notes associated with a character
    get_notes_page_by_character_id(character_id: int, page: int, per_page: int)
        Get a single page of notes associated with a character from the database
    append_images_to_character(character_id: int, images: list)
        Append images to a character
    change_image_position(image_id: int, position: int)
        Set the position of an image related to a character
    change_image_default_status(image_id: int, is_default: bool)
        Set the default status of an image related to a character
    delete_image(image_id: int)
        Delete an image related to a character
    get_image_count_by_character_id(character_id: int)
        Get image count associated with a character
    get_images_by_character_id(character_id: int)
        Get all images associated with a character
    get_images_page_by_character_id(character_id: int, page: int, per_page: int)
        Get a single page of images associated with a character from the database
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_character(
        self, title: str = None, honorific: str = None, first_name: str = None,
        middle_name: str = None, last_name: str = None, nickname: str = None,
        gender: str = None, sex: str = None, ethnicity: str = None,
        race: str = None, tribe_or_clan: str = None, nationality: str = None,
        religion: str = None, occupation: str = None, education: str = None,
        marital_status: str = None, children: bool = None,
        date_of_birth: str = None, date_of_death: str = None,
        description: str = None, mbti: str = None, enneagram: str = None,
        wounds: str = None
    ) -> Character:
        """Create a new character

        Although all the attributes are technically optional, at least one of the following fields must be provided:
        title, first name, middle name, last name, or nickname.

        Parameters
        ----------
        title : str
            The title of the character, optional
        honorific : str
            The honorific of the character, optional
        first_name : str
            The first name of the character, optional
        middle_name : str
            The middle name of the character, optional
        last_name : str
            The last name of the character, optional
        nickname : str
            The nickname of the character, optional
        gender: str
            The gender of the character, optional
        sex: str
            The sex of the character, optional
        ethnicity: str
            The ethnicity of the character, optional
        race: str
            The race of the character, optional
        tribe_or_clan: str
            The tribe or clan of the character, optional
        nationality: str
            The nationality of the character, optional
        religion: str
            The religion of the character, optional
        occupation: str
            The occupation of the character, optional
        education: str
            The education of the character, optional
        marital_status: str
            The marital status of the character, optional
        children: bool
            Whether the character has children, optional
        date_of_birth: str
            The date of birth of the character, optional
        date_of_death: str
            The date of death of the character, optional
        description
            The description of the character, optional
        mbti: str
            The Myers-Briggs Type Indicator of the character, optional
        enneagram: str
            The Enneagram type of the character, optional
        wounds: str
            The wounds of the character, optional

        Returns
        -------
        Character
            The new character object
        """

        with self._session as session:

            try:

                if not title and not first_name and not last_name and not middle_name and not nickname:
                    raise ValueError("""At least one of the following fields 
                    must be provided: title, first name, middle name, last name, 
                    nickname.""")

                created = datetime.now()
                modified = created
                character = Character(
                    user_id=self._owner.id, title=title, first_name=first_name,
                    middle_name=middle_name, last_name=last_name,
                    honorific=honorific, nickname=nickname, gender=gender,
                    sex=sex, ethnicity=ethnicity, race=race,
                    tribe_or_clan=tribe_or_clan, nationality=nationality,
                    religion=religion, occupation=occupation,
                    education=education, marital_status=marital_status,
                    children=children, date_of_birth=date_of_birth,
                    date_of_death=date_of_death, description=description,
                    mbti=mbti, enneagram=enneagram, wounds=wounds,
                    created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'{character.__str__} \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(character)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character

    def update_character(
        self, character_id: int = None, title: str = None, honorific: str = None,
        first_name: str = None, middle_name: str = None, last_name: str = None,
        nickname: str = None, gender: str = None, sex: str = None,
        ethnicity: str = None, race: str = None, tribe_or_clan: str = None,
        nationality: str = None, religion: str = None, occupation: str = None,
        education: str = None, marital_status: str = None, children: bool = None,
        date_of_birth: str = None, date_of_death: str = None,
        description: str = None, mbti: str = None, enneagram: str = None,
        wounds: str = None
    ) -> Type[Character]:
        """Update a character

        Parameters
        ----------
        character_id : int
            The id of the character to update
        title : str
            The title of the character
        honorific : str
            The honorific of the character
        first_name : str
            The first name of the character
        middle_name : str
            The middle name of the character
        last_name : str
            The last name of the character
        nickname : str
            The nickname of the character
        gender : str
            The gender of the character
        sex : str
            The sex of the character
        ethnicity : str
            The ethnicity of the character
        race : str
            The race of the character
        tribe_or_clan : str
            The tribe or clan of the character
        nationality : str
            The nationality of the character
        religion : str
            The religion of the character
        occupation : str
            The occupation of the character
        education : str
            The education of the character
        marital_status : str
            The marital status of the character
        children : bool
            Whether the character has children
        date_of_birth : str
            The date of birth of the character
        date_of_death : str
            The date of death of the character
        description : str
            The description of the character
        mbti : str
            The Myers-Briggs Type Indicator of the character
        enneagram : str
            The Enneagram type of the character
        wounds : str
            The wounds of the character

        Returns
        -------
        Character
            The updated character object
        """

        with self._session as session:

            try:

                character = session.query(Character).filter(
                    Character.id == character_id,
                    Character.user_id == self._owner.id
                ).first()

                if not character:
                    raise ValueError('Character not found.')

                character.title = title
                character.honorific = honorific
                character.first_name = first_name
                character.middle_name = middle_name
                character.last_name = last_name
                character.nickname = nickname
                character.gender = gender
                character.sex = sex
                character.ethnicity = ethnicity
                character.race = race
                character.tribe_or_clan = tribe_or_clan
                character.nationality = nationality
                character.religion = religion
                character.occupation = occupation
                character.education = education
                character.marital_status = marital_status
                character.children = children
                character.date_of_birth = date_of_birth
                character.date_of_death = date_of_death
                character.description = description
                character.mbti = mbti
                character.enneagram = enneagram
                character.wounds = wounds
                character.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'{character.__str__} \
                    updated by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character

    def delete_character(self, character_id: int) -> bool:
        """Delete a character

        Character objects are even more complex than Chapter objects. Each Character can have an arbitrary number of
        Link, Note, and Image objects associated with it. Additionally, each Character has an arbitrary number of
        character relationships, an arbitrary number of character traits, and an arbitrary number of events associated
        with it. Before the Character can be deleted, the links, notes, images, relationships, traits, and events
        associated with it must be deleted. The Characters do not possess a position scheme, so there is no need to
        adjust the position of sibling characters. Finally, the Character is deleted.

        Parameters
        ----------
        character_id : int
            The id of the character to delete

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:

            try:

                character = session.query(Character).filter(
                    Character.id == character_id,
                    Character.user_id == self._owner.id
                ).first()

                if not character:
                    raise ValueError('Character not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'{character.__str__} \
                    deleted by {self._owner.username}', created=datetime.now()
                )

                session.delete(character)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_character_by_id(self, character_id: int) -> Type[Character] | None:
        """Get a character by id

        Parameters
        ----------
        character_id : int
            The id of the character to get

        Returns
        -------
        Type[Character] | None
            The character if found, otherwise None
        """

        with self._session as session:
            character = session.query(Character).filter(
                Character.id == character_id,
                Character.user_id == self._owner.id
            ).first()
            return character if character else None

    def get_character_count(self) -> int:
        """Get character count associated with a user

        Returns
        -------
        int
            The count of characters associated with the user
        """

        with self._session as session:
            return session.query(func.count(Character.id)).filter(
                Character.user_id == self._owner.id
            ).scalar()

    def get_all_characters(self) -> List[Type[Character]]:
        """Get all characters associated with a user

        Returns
        -------
        list
            The list of characters associated with the user
        """

        with self._session as session:
            return session.query(Character).filter(
                Character.user_id == self._owner.id
            ).all()

    def get_all_characters_page(
        self, page: int, per_page: int
    ) -> List[Type[Character]]:
        """Get a single page of characters from the database associated with a user

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            The list of characters associated with the user
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Character).filter(
                Character.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def get_character_count_by_story_id(self, story_id: int) -> int:
        """Get character count associated with a story

        The characters and stories are associated in the CharacterStory table.

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        int
            The count of characters associated with the story
        """

        with self._session as session:
            return session.query(
                func.count(CharacterStory.character_id)
            ).filter(
                CharacterStory.story_id == story_id,
                CharacterStory.user_id == self._owner.id
            ).scalar()

    def get_characters_by_story_id(
        self, story_id: int
    ) -> List[Type[Character]]:
        """Get characters by story id

        The characters and stories are associated in the CharacterStory table.

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            The list of characters associated with the story
        """

        with self._session as session:
            for character_story in session.query(CharacterStory).filter(
                CharacterStory.story_id == story_id,
                CharacterStory.user_id == self._owner.id
            ).all():
                yield session.query(Character).filter(
                    Character.id == character_story.character_id,
                    Character.user_id == self._owner.id
                ).first()

    def get_characters_page_by_story_id(
        self, story_id: int, page: int, per_page: int
    ) -> List[Type[Character]]:
        """Get a single page of characters by story id

        The characters and stories are associated in the CharacterStory table.

        Parameters
        ----------
        story_id : int
            The id of the story
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            The list of characters associated with the story
        """

        with self._session as session:
            offset = (page - 1) * per_page
            for character_story in session.query(CharacterStory).filter(
                CharacterStory.story_id == story_id,
                    CharacterStory.user_id == self._owner.id
            ).offset(offset).limit(per_page).all():
                yield session.query(Character).filter(
                    Character.id == character_story.character_id,
                    Character.user_id == self._owner.id
                ).first()

    def search_characters(self, search: str) -> List[Type[Character]]:
        """Search for characters by title, first name, middle name, last name, nickname, and description belonging to \
        a specific user

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            The list of characters associated with the user
        """

        with self._session as session:
            return session.query(Character).filter(
                or_(
                    Character.title.like(f'%{search}%'),
                    Character.first_name.like(f'%{search}%'),
                    Character.middle_name.like(f'%{search}%'),
                    Character.last_name.like(f'%{search}%'),
                    Character.nickname.like(f'%{search}%'),
                    Character.description.like(f'%{search}%')
                ),
                Character.user_id == self._owner.id
            ).all()

    def search_characters_by_story_id(
        self, story_id: int, search: str
    ) -> List[Type[Character]]:
        """Search for characters by title, first name, middle name, last name, nickname, and description belonging to \
        a specific story

        Parameters
        ----------
        story_id : int
            The id of the story
        search : str
            The search string

        Returns
        -------
        list
            The list of characters associated with the story
        """

        with self._session as session:
            return session.query(Character).join(
                CharacterStory,
                Character.id == CharacterStory.character_id
            ).filter(
                or_(
                    Character.title.like(f'%{search}%'),
                    Character.first_name.like(f'%{search}%'),
                    Character.middle_name.like(f'%{search}%'),
                    Character.last_name.like(f'%{search}%'),
                    Character.nickname.like(f'%{search}%'),
                    Character.description.like(f'%{search}%')
                ),
                CharacterStory.story_id == story_id,
                CharacterStory.user_id == self._owner.id
            ).all()

    def create_relationship(
        self, parent_id: int, related_id: int, relationship_type: str,
        description: str = None, start_date: str = None, end_date: str = None
    ) -> CharacterRelationship:
        """Create a new character relationship

        There are many relationships for each character, and the linking class, CharacterRelationship, is used to
        define the relationship between two characters. The relationship type is defined in the
        CharacterRelationshipTypes and elaborated upon in the description. Before insertion, in order to determine
        the correct position, the highest position among sibling CharacterRelationship objects is first determined.
        The new position is set to the highest position plus one. The created and modified fields are set to the
        current date and time. Finally, insert the new CharacterRelationship object into the database.

        Parameters
        ----------
        parent_id : int
            The id of the parent character
        related_id : int
            The id of the related character
        relationship_type : str
            The type of relationship
        description : str
            The description of the relationship, optional
        start_date : str
            The start date of the relationship, optional
        end_date : str
            The end date of the relationship, optional

        Returns
        -------
        CharacterRelationship
            The new character relationship object
        """

        with self._session as session:
            try:
                if not parent_id or not related_id:
                    raise ValueError('Both parent and related character ids must be provided.')

                if parent_id == related_id:
                    raise ValueError('Parent and related character ids must be different.')

                parent = session.query(Character).filter(
                    Character.id == parent_id, Character.user_id == self._owner.id
                ).first()

                if not parent:
                    raise ValueError('Parent character not found.')

                related = session.query(Character).filter(
                    Character.id == related_id, Character.user_id == self._owner.id
                ).first()

                if not related:
                    raise ValueError('Related character not found.')

                position = session.query(func.max(CharacterRelationship.position)).filter(
                    CharacterRelationship.parent_id == parent_id,
                    CharacterRelationship.user_id == self._owner.id
                ).scalar()

                position = 1 if not position else position + 1
                created = datetime.now()
                modified = created
                character_relationship = CharacterRelationship(
                    user_id=self._owner.id, parent_id=parent.id,
                    related_id=related.id, position=position,
                    relationship_type=relationship_type,
                    description=description, start_date=start_date,
                    end_date=end_date, created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character relationship \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(character_relationship)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character_relationship

    def update_relationship(
        self, relationship_id: int, parent_id: int, related_id: int,
        relationship_type: str, description: str = None, start_date: str = None,
        end_date: str = None
    ) -> Type[CharacterRelationship]:
        """Update a character relationship

        Parameters
        ----------
        relationship_id : int
            The id of the relationship to update
        parent_id : int
            The id of the parent character
        related_id : int
            The id of the related character
        relationship_type : str
            The type of relationship
        description : str
            The description of the relationship, optional
        start_date : str
            The start date of the relationship, optional
        end_date : str
            The end date of the relationship, optional

        Returns
        -------
        CharacterRelationship
            The updated character relationship object
        """

        with self._session as session:
            try:
                character_relationship = session.query(
                    CharacterRelationship
                ).filter(
                    CharacterRelationship.id == relationship_id,
                    CharacterRelationship.user_id == self._owner.id
                ).first()

                if not character_relationship:
                    raise ValueError('Character relationship not found.')

                character_relationship.parent_id = parent_id
                character_relationship.related_id = related_id
                character_relationship.relationship_type = relationship_type
                character_relationship.description = description
                character_relationship.start_date = start_date
                character_relationship.end_date = end_date
                character_relationship.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character relationship \
                    updated by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character_relationship

    def change_relationship_position(
        self, relationship_id: int, position: int
    ) -> Type[CharacterRelationship]:
        """Set the position of a character relationship

        First, determine whether the new position is closer to 1 or further from 1. If closer to one, get all sibling
        CharacterRelationship objects with positions greater than or equal to the new position but less than the current
        position, and increment those position values by 1. If the target position is further away from 1 than the
        current position, get all sibling CharacterRelationship objects with positions greater than the current position
        but less than or equal to the new position, and decrement those position values by 1. Finally, set the position
        of the target CharacterRelationship object to the new position. Return the new position.

        Parameters
        ----------
        relationship_id : int
            The id of the relationship to update
        position : int
            The new position of the relationship

        Returns
        -------
        CharacterRelationship
            The updated character relationship object
        """

        with self._session as session:
            try:
                character_relationship = session.query(CharacterRelationship).filter(
                    CharacterRelationship.id == relationship_id, CharacterRelationship.user_id == self._owner.id
                ).first()

                if not character_relationship:
                    raise ValueError('Character relationship not found.')

                if position < 1:
                    raise ValueError('Position must be greater than 0.')

                if position == character_relationship.position:
                    return character_relationship

                if position < character_relationship.position:
                    for sibling in session.query(CharacterRelationship).filter(
                        CharacterRelationship.parent_id == character_relationship.parent_id,
                        CharacterRelationship.position >= position,
                        CharacterRelationship.position < character_relationship.position,
                        CharacterRelationship.user_id == self._owner.id
                    ).all():
                        sibling.position += 1

                else:
                    for sibling in session.query(CharacterRelationship).filter(
                        CharacterRelationship.parent_id == character_relationship.parent_id,
                        CharacterRelationship.position > character_relationship.position,
                        CharacterRelationship.position <= position,
                        CharacterRelationship.user_id == self._owner.id
                    ).all():
                        sibling.position -= 1

                character_relationship.position = position
                character_relationship.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character relationship \
                    position updated by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character_relationship

    def delete_relationship(self, relationship_id: int) -> bool:
        """Delete a character relationship

        Before deleting the CharacterRelationship object, all sibling objects with positions higher than the current
        position are decremented by one. The activity is then logged and the CharacterRelationship object is deleted.

        Parameters
        ----------
        relationship_id : int
            The id of the relationship to delete

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                character_relationship = session.query(
                    CharacterRelationship
                ).filter(
                    CharacterRelationship.id == relationship_id,
                    CharacterRelationship.user_id == self._owner.id
                ).first()

                if not character_relationship:
                    raise ValueError('Character relationship not found.')

                for sibling in session.query(CharacterRelationship).filter(
                    CharacterRelationship.parent_id == character_relationship.parent_id,
                    CharacterRelationship.position > character_relationship.position,
                    CharacterRelationship.user_id == self._owner.id
                ).all():
                    sibling.position -= 1

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character relationship \
                    deleted by {self._owner.username}', created=datetime.now()
                )

                session.delete(character_relationship)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_relationship_by_id(
        self, relationship_id: int
    ) -> Type[CharacterRelationship] | None:
        """Get a character relationship by id

        Parameters
        ----------
        relationship_id : int
            The id of the relationship to get

        Returns
        -------
        Type[CharacterRelationship] | None
            The character relationship if found, otherwise None
        """

        with self._session as session:
            character_relationship = session.query(
                CharacterRelationship
            ).filter(
                CharacterRelationship.id == relationship_id,
                CharacterRelationship.user_id == self._owner.id
            ).first()
            return character_relationship if character_relationship else None

    def get_relationships_by_character_id(
        self, parent_id: int
    ) -> List[Type[CharacterRelationship]]:
        """Get character relationships by character id, from that character's perspective

        Parameters
        ----------
        parent_id : int
            The id of the character

        Returns
        -------
        list
            The list of character relationships
        """

        with self._session as session:
            return session.query(CharacterRelationship).filter(
                CharacterRelationship.parent_id == parent_id,
                CharacterRelationship.user_id == self._owner.id
            ).all()

    def get_relationships_page_by_character_id(
        self, parent_id: int, page: int, per_page: int
    ) -> List[Type[CharacterRelationship]]:
        """Get a single page of character relationships by character id, from that character's perspective

        Parameters
        ----------
        parent_id : int
            The id of the character
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            The list of character relationships
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(CharacterRelationship).filter(
                CharacterRelationship.parent_id == parent_id,
                CharacterRelationship.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def create_trait(
        self, character_id: int, name: str, magnitude: int
    ) -> CharacterTrait:
        """Create a new character trait

        The position of the new trait must be determined by first finding the highest existing position among siblings,
        if any. Any other CharacterTrait objects associated with the same character are those siblings.

        Parameters
        ----------
        character_id : int
            The id of the character
        name : str
            The name of the trait
        magnitude : int
            The magnitude of the trait

        Returns
        -------
        CharacterTrait
            The new character trait object
        """

        with self._session as session:
            try:
                if not character_id or not name or not magnitude:
                    raise ValueError('Character id, name, and magnitude must be provided.')

                character = session.query(Character).filter(
                    Character.id == character_id,
                    Character.user_id == self._owner.id
                ).first()

                if not character:
                    raise ValueError('Character not found.')

                position = session.query(
                    func.max(CharacterTrait.position)
                ).filter(
                    CharacterTrait.character_id == character_id,
                    CharacterTrait.user_id == self._owner.id
                ).scalar()

                position = 1 if not position else position + 1
                created = datetime.now()
                modified = created
                character_trait = CharacterTrait(
                    user_id=self._owner.id, character_id=character_id,
                    position=position, name=name, magnitude=magnitude,
                    created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character trait {name} \
                    created by {self._owner.username} for "{character.__str__}"',
                    created=datetime.now()
                )

                session.add(character_trait)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character_trait

    def update_trait(
        self, trait_id: int, name: str, magnitude: int
    ) -> Type[CharacterTrait]:
        """Update a character trait

        Parameters
        ----------
        trait_id : int
            The id of the trait to update
        name : str
            The name of the trait
        magnitude : int
            The magnitude of the trait

        Returns
        -------
        CharacterTrait
            The updated character trait object
        """

        with self._session as session:
            try:
                character_trait = session.query(CharacterTrait).filter(
                    CharacterTrait.id == trait_id,
                    CharacterTrait.user_id == self._owner.id
                ).first()

                if not character_trait:
                    raise ValueError('Character trait not found.')

                character_trait.name = name
                character_trait.magnitude = magnitude
                character_trait.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character trait \
                    {character_trait.__str__} updated by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character_trait

    def change_trait_position(
        self, trait_id: int, position: int
    ) -> Type[CharacterTrait]:
        """Set the position of a character trait

        First, determine whether the new position is closer to 1 or further from 1. If closer to one, get all sibling
        CharacterTraits with positions greater than or equal to the new position but less than the current position, and
        increment those position values by 1. If the target position is further away from 1 than the current position,
        get all sibling CharacterTraits with positions greater than the current position but less than or equal to the
        new position, and decrement those position values by 1. Finally, set the position of the target chapter to the
        new position. Return the new position.

        Parameters
        ----------
        trait_id : int
            The id of the trait to update
        position : int
            The position of the trait

        Returns
        -------
        CharacterTrait
            The updated character trait object
        """

        with self._session as session:
            try:

                config = ConfigParser()
                config.read("config.cfg")
                datetime_format = config.get("formats", "datetime")
                character_trait = session.query(CharacterTrait).filter(
                    CharacterTrait.id == trait_id, CharacterTrait.user_id == self._owner.id
                ).first()

                if not character_trait:
                    raise ValueError('Character trait not found.')

                if position < 1:
                    raise ValueError('Position must be greater than 0.')

                highest_position = session.query(func.max(CharacterTrait.position)).filter(
                    CharacterTrait.character_id == character_trait.character_id,
                    CharacterTrait.user_id == self._owner.id
                ).scalar()

                if position > highest_position:
                    raise ValueError(f'Position must be less than or equal to {highest_position}.')

                if position == character_trait.position:
                    return character_trait.position

                if position < character_trait.position:
                    siblings = session.query(CharacterTrait).filter(
                        CharacterTrait.character_id == character_trait.character_id,
                        CharacterTrait.position >= position,
                        CharacterTrait.user_id == self._owner.id,
                        CharacterTrait.position < character_trait.position
                    ).all()

                    for sibling in siblings:
                        sibling.position += 1
                        sibling.created = datetime.strptime(
                            str(sibling.created), datetime_format
                        )
                        sibling.modified = datetime.now()

                else:
                    siblings = session.query(CharacterTrait).filter(
                        CharacterTrait.character_id == character_trait.character_id,
                        CharacterTrait.position > character_trait.position,
                        CharacterTrait.position <= position,
                        CharacterTrait.user_id == self._owner.id
                    ).all()

                    for sibling in siblings:
                        sibling.position -= 1
                        sibling.created = datetime.strptime(
                            str(sibling.created), datetime_format
                        )
                        sibling.modified = datetime.now()

                character_trait.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character trait \
                    {character_trait.__str__} position changed by \
                    {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character_trait

    def delete_trait(self, trait_id: int) -> bool:
        """Delete a character trait

        When a trait is deleted, all sibling traits having a higher position than the deleted trait must have their
        positions each decremented by 1. The activity is then logged and the CharacterTrait object is deleted.

        Parameters
        ----------
        trait_id : int
            The id of the trait to delete

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:

                config = ConfigParser()
                config.read("config.cfg")
                datetime_format = config.get("formats", "datetime")

                character_trait = session.query(CharacterTrait).filter(
                    CharacterTrait.id == trait_id,
                    CharacterTrait.user_id == self._owner.id
                ).first()

                if not character_trait:
                    raise ValueError('Character trait not found.')

                for sibling in session.query(CharacterTrait).filter(
                    CharacterTrait.character_id == character_trait.character_id,
                    CharacterTrait.position > character_trait.position,
                    CharacterTrait.user_id == self._owner.id
                ).all():
                    sibling.position -= 1
                    sibling.created = datetime.strptime(
                        str(sibling.created), datetime_format
                    )
                    sibling.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character trait \
                    {character_trait.__str__} deleted by \
                    {self._owner.username}', created=datetime.now()
                )

                session.delete(character_trait)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_trait_by_id(self, trait_id: int) -> Type[CharacterTrait] | None:
        """Get a character trait by id

        Parameters
        ----------
        trait_id : int
            The id of the trait to get

        Returns
        -------
        Type[CharacterTrait] | None
            The character trait if found, otherwise None
        """

        with self._session as session:
            character_trait = session.query(CharacterTrait).filter(
                CharacterTrait.id == trait_id,
                CharacterTrait.user_id == self._owner.id
            ).first()
            return character_trait if character_trait else None

    def get_traits_by_character_id(
        self, character_id: int
    ) -> List[Type[CharacterTrait]]:
        """Get character traits by character id

        Parameters
        ----------
        character_id : int
            The id of the character

        Returns
        -------
        list
            The list of character traits
        """

        with self._session as session:
            return session.query(CharacterTrait).filter(
                CharacterTrait.character_id == character_id,
                CharacterTrait.user_id == self._owner.id
            ).all()

    def get_traits_page_by_character_id(
        self, character_id: int, page: int, per_page: int
    ) -> List[Type[CharacterTrait]]:
        """Get a single page of character traits by character id

        Parameters
        ----------
        character_id : int
            The id of the character
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            The list of character traits
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(CharacterTrait).filter(
                CharacterTrait.character_id == character_id,
                CharacterTrait.user_id == self._owner.id
            ).offset(
                offset).limit(per_page).all()

    def append_events_to_character(
        self, character_id: int, event_ids: list
    ) -> Type[Character]:
        """Append events to a character

        Parameters
        ----------
        character_id : int
            The id of the character
        event_ids : list
            A list of event ids

        Returns
        -------
        Character
            The updated character object
        """

        with self._session as session:
            try:
                character = session.query(Character).filter(
                    Character.id == character_id,
                    Character.user_id == self._owner.id
                ).first()

                if not character:
                    raise ValueError('Character not found.')

                for event_id in event_ids:
                    event = session.query(Event).filter(
                        Event.id == event_id, Event.user_id == self._owner.id
                    ).first()

                    if not event:
                        raise ValueError('Event not found.')

                    character_event = CharacterEvent(
                        user_id=self._owner.id, character_id=character_id,
                        event_id=event_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Event \
                        {event.title[:50]} associated with {character.__str__} \
                        by {self._owner.username}', created=datetime.now()
                    )

                    session.add(character_event)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character

    def get_events_by_character_id(
        self, character_id: int
    ) -> List[Type[Event]]:
        """Get all events associated with a character

        Using the association of the CharacterEvent table, get all events from teh Event table associated with a
        character.

        Parameters
        ----------
        character_id : int
            The id of the character

        Returns
        -------
        list
            The list of events
        """

        with self._session as session:
            for character_event in session.query(CharacterEvent).filter(
                CharacterEvent.character_id == character_id,
                    CharacterEvent.user_id == self._owner.id
            ).all():
                yield session.query(Event).filter(
                    Event.id == character_event.event_id,
                    Event.user_id == self._owner.id
                ).first()

    def get_events_page_by_character_id(
        self, character_id: int, page: int, per_page: int
    ) -> List[Type[Event]]:
        """Get a single page of events associated with a character from the database

        Parameters
        ----------
        character_id : int
            The id of the character
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            The list of events
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(CharacterEvent).filter(
                CharacterEvent.character_id == character_id,
                CharacterEvent.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def append_links_to_character(
        self, character_id: int, link_ids: list
    ) -> Type[Character]:
        """Append links to a character

        Parameters
        ----------
        character_id : int
            The id of the character
        link_ids : list
            A list of link ids

        Returns
        -------
        Character
            The updated character object
        """

        with self._session as session:
            try:
                character = session.query(Character).filter(
                    Character.id == character_id,
                    Character.user_id == self._owner.id
                ).first()

                if not character:
                    raise ValueError('Character not found.')

                for link_id in link_ids:
                    link = session.query(Link).filter(
                        Link.id == link_id,
                        Link.user_id == self._owner.id
                    ).first()

                    if not link:
                        raise ValueError('Link not found.')

                    character_link = CharacterLink(
                        user_id=self._owner.id, character_id=character_id,
                        link_id=link_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Link \
                        {link.title[:50]} associated with {character.__str__} \
                        by {self._owner.username}', created=datetime.now()
                    )

                    session.add(character_link)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character

    def get_links_by_character_id(self, character_id: int) -> List[Type[Link]]:
        """Get all links associated with a character

        Parameters
        ----------
        character_id : int
            The id of the character

        Returns
        -------
        list
            The list of links
        """

        with self._session as session:
            for character_link in session.query(CharacterLink).filter(
                CharacterLink.character_id == character_id,
                CharacterLink.user_id == self._owner.id
            ).all():
                yield session.query(Link).filter(
                    Link.id == character_link.link_id,
                    Link.user_id == self._owner.id
                ).first()

    def get_links_page_by_character_id(
        self, character_id: int, page: int, per_page: int
    ) -> List[Type[Link]]:
        """Get a single page of links associated with a character from the database

        Parameters
        ----------
        character_id : int
            The id of the character
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            The list of links
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(CharacterLink).filter(
                CharacterLink.character_id == character_id,
                CharacterLink.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def append_notes_to_character(
        self, character_id: int, note_ids: list
    ) -> Type[Character]:
        """Append notes to a character

        Parameters
        ----------
        character_id : int
            The id of the character
        note_ids : list
            A list of note ids

        Returns
        -------
        Character
            The updated character object
        """

        with self._session as session:
            try:
                character = session.query(Character).filter(
                    Character.id == character_id,
                    Character.user_id == self._owner.id
                ).first()

                if not character:
                    raise ValueError('Character not found.')

                for note_id in note_ids:
                    note = session.query(Note).filter(
                        Note.id == note_id,
                        Note.user_id == self._owner.id
                    ).first()

                    if not note:
                        raise ValueError('Note not found.')

                    character_note = CharacterNote(
                        user_id=self._owner.id, character_id=character_id,
                        note_id=note_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Note \
                        {note.title[:50]} associated with {character.__str__} \
                        by {self._owner.username}', created=datetime.now()
                    )

                    session.add(character_note)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character

    def get_notes_by_character_id(self, character_id: int) -> List[Type[Note]]:
        """Get all notes associated with a character

        Parameters
        ----------
        character_id : int
            The id of the character

        Returns
        -------
        list
            The list of notes
        """

        with self._session as session:
            for character_note in session.query(CharacterNote).filter(
                CharacterNote.character_id == character_id,
                CharacterNote.user_id == self._owner.id
            ).all():
                yield session.query(Note).filter(
                    Note.id == character_note.note_id,
                    Note.user_id == self._owner.id
                ).first()

    def get_notes_page_by_character_id(
        self, character_id: int, page: int, per_page: int
    ) -> List[Type[Note]]:
        """Get a single page of notes associated with a character from the database

        Parameters
        ----------
        character_id : int
            The id of the character
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            The list of notes
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(CharacterNote).filter(
                CharacterNote.character_id == character_id,
                CharacterNote.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def append_images_to_character(
        self, character_id: int, image_ids: list
    ) -> Type[Character]:
        """Append images to a character

        As images are appended to the character, before each image is appended, the highest position among other images
        associated with this object is found and the new image is appended with a position one higher than the highest
        position.

        Parameters
        ----------
        character_id : int
            The id of the character
        image_ids : list
            A list of image ids

        Returns
        -------
        Character
            The updated character object
        """

        with self._session as session:
            try:
                character = session.query(Character).filter(
                    Character.id == character_id,
                    Character.user_id == self._owner.id
                ).first()

                if not character:
                    raise ValueError('Character not found.')

                for image_id in image_ids:
                    image = session.query(Image).filter(
                        Image.id == image_id,
                        Image.user_id == self._owner.id
                    ).first()

                    if not image:
                        raise ValueError('Image not found.')

                    position = session.query(
                        func.max(CharacterImage.position)
                    ).filter(
                        CharacterImage.character_id == character_id,
                        CharacterImage.user_id == self._owner.id
                    ).scalar()

                    position = 1 if not position else position + 1
                    is_default = False
                    created = datetime.now()
                    modified = created
                    character_image = CharacterImage(
                        user_id=self._owner.id, character_id=character_id,
                        image_id=image_id, position=position,
                        is_default=is_default, created=created,
                        modified=modified
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Image \
                        {image.filename[:50]} associated with character \
                        {str(character)[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(character_image)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character

    def change_image_position(
        self, image_id: int, position: int
    ) -> Type[CharacterImage]:
        """Set the position of a character image

        First, determine whether the new position is closer to 1 or further from 1. If closer to one, get all sibling
        CharacterImage objects with positions greater than or equal to the new position but less than the current
        position, and increment those position values by 1. If the target position is further away from 1 than the
        current position, get all sibling CharacterImage objects with positions greater than the current position but
        less than or equal to the new position, and decrement those position values by 1. Finally, set the position of
        the target CharacterImage object to the new position. Return the new position.

        Parameters
        ----------
        image_id : int
            The id of the image to update
        position : int
            The position of the image

        Returns
        -------
        CharacterImage
            The updated character image object
        """

        with self._session as session:
            try:

                config = ConfigParser()
                config.read("config.cfg")
                datetime_format = config.get("formats", "datetime")

                character_image = session.query(CharacterImage).filter(
                    CharacterImage.id == image_id,
                    CharacterImage.user_id == self._owner.id
                ).first()

                if not character_image:
                    raise ValueError('Character image not found.')

                if position < 1:
                    raise ValueError('Position must be greater than 0.')

                highest_position = session.query(
                    func.max(CharacterImage.position)
                ).filter(
                    CharacterImage.character_id == character_image.character_id,
                    CharacterImage.user_id == self._owner.id
                ).scalar()

                if position > highest_position:
                    raise ValueError(f'Position must be less than or equal to {highest_position}.')

                if position == character_image.position:
                    return character_image.position

                if position < character_image.position:
                    siblings = session.query(CharacterImage).filter(
                        CharacterImage.character_id == character_image.character_id,
                        CharacterImage.position >= position,
                        CharacterImage.user_id == self._owner.id,
                        CharacterImage.position < character_image.position
                    ).all()

                    for sibling in siblings:
                        sibling.position += 1
                        sibling.created = datetime.strptime(
                            str(sibling.created), datetime_format
                        )
                        sibling.modified = datetime.now()

                else:
                    siblings = session.query(CharacterImage).filter(
                        CharacterImage.character_id == character_image.character_id,
                        CharacterImage.position > character_image.position,
                        CharacterImage.position <= position,
                        CharacterImage.user_id == self._owner.id
                    ).all()

                    for sibling in siblings:
                        sibling.position -= 1
                        sibling.created = datetime.strptime(
                            str(sibling.created), datetime_format
                        )
                        sibling.modified = datetime.now()

                character_image.position = position
                character_image.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character image \
                    {character_image.__str__} position changed by \
                    {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character_image

    def change_image_default_status(
        self, image_id: int, is_default: bool
    ) -> Type[CharacterImage]:
        """Set the default status of a character image

        If changing the value from true to false, it is straightforward - make the value change and save the object.
        However, if the value is changed from false to true, then before saving that new value in the object, the same
        attribute must first be set to false in all sibling CharacterImage objects. This is because only one image can
        be the default image for a character. The activity is then logged and the CharacterImage object is saved.

        Parameters
        ----------
        image_id : int
            The id of the image to update
        is_default : bool
            The default status of the image

        Returns
        -------
        CharacterImage
            The updated character image object
        """

        with self._session as session:
            try:
                character_image = session.query(CharacterImage).filter(
                    CharacterImage.id == image_id,
                    CharacterImage.user_id == self._owner.id
                ).first()

                if not character_image:
                    raise ValueError('Character image not found.')

                if is_default == character_image.is_default:
                    return character_image

                if is_default:
                    for sibling in session.query(CharacterImage).filter(
                        CharacterImage.character_id == character_image.character_id,
                        CharacterImage.user_id == self._owner.id
                    ).all():
                        sibling.is_default = False
                        sibling.modified = datetime.now()

                character_image.is_default = is_default
                character_image.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character image \
                    {character_image.__str__} default status changed by \
                    {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return character_image

    def delete_image(self, image_id: int) -> bool:
        """Delete a character image

        When an image is deleted, all sibling images having a higher position than the deleted image must have their
        positions each decremented by 1. The activity is then logged and the CharacterImage object is deleted. The
        Image object is also deleted.

        Parameters
        ----------
        image_id : int
            The id of the image to delete

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:

                config = ConfigParser()
                config.read("config.cfg")
                datetime_format = config.get("formats", "datetime")

                character_image = session.query(CharacterImage).filter(
                    CharacterImage.id == image_id,
                    CharacterImage.user_id == self._owner.id
                ).first()
                image = session.query(Image).filter(
                    Image.id == character_image.image_id,
                    Image.user_id == self._owner.id
                ).first()

                if not character_image:
                    raise ValueError('Character image not found.')

                for sibling in session.query(CharacterImage).filter(
                    CharacterImage.character_id == character_image.character_id,
                    CharacterImage.position > character_image.position,
                        CharacterImage.user_id == self._owner.id
                ).all():
                    sibling.position -= 1
                    sibling.created = datetime.strptime(
                        str(sibling.created), datetime_format
                    )
                    sibling.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character image \
                    {image.caption[:50]} deleted by {self._owner.username}',
                    created=datetime.now()
                )

                session.delete(character_image)
                session.delete(image)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_image_count_by_character_id(self, character_id: int) -> int:
        """Get image count associated with a character

        Parameters
        ----------
        character_id : int
            The id of the character

        Returns
        -------
        int
            The count of images
        """

        with self._session as session:
            return session.query(func.count(CharacterImage.id)).filter(
                CharacterImage.character_id == character_id,
                CharacterImage.user_id == self._owner.id
            ).scalar()

    def get_images_by_character_id(
        self, character_id: int
    ) -> List[Type[Image]]:
        """Get all images associated with a character

        The images will be returned in the order of their position. A yield is used to return the images one at a time.

        Parameters
        ----------
        character_id : int
            The id of the character

        Returns
        -------
        list
            The list of images
        """

        with self._session as session:
            for character_image in session.query(CharacterImage).filter(
                CharacterImage.character_id == character_id,
                    CharacterImage.user_id == self._owner.id
            ).order_by(CharacterImage.position).all():
                yield session.query(Image).filter(
                    Image.id == character_image.image_id,
                    Image.user_id == self._owner.id
                ).first()

    def get_images_page_by_character_id(
        self, character_id: int, page: int, per_page: int
    ) -> List[Type[Image]]:
        """Get a single page of images associated with a character from the database

        The images will be returned in the order of their position. A yield is used to return the images one at a time.

        Parameters
        ----------
        character_id : int
            The id of the character
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            The list of images
        """

        with self._session as session:
            offset = (page - 1) * per_page
            for character_image in session.query(CharacterImage).filter(
                CharacterImage.character_id == character_id,
                    CharacterImage.user_id == self._owner.id
            ).order_by(
                CharacterImage.position
            ).offset(offset).limit(per_page).all():
                yield session.query(Image).filter(
                    Image.id == character_image.image_id,
                    Image.user_id == self._owner.id
                ).first()



from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, String, Text, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import Scene, User, Story, ChapterLink, ChapterNote, Base


class Chapter(Base):
    """The Chapter class represents a chapter of a story.

    Attributes
    ----------
        id: int
            The chapter's id
        user_id: int
            The id of the owner of this entry
        story_id: int
            The story's id
        position: int
            The chapter's position in the story
        title: str
            The chapter's title
        description: str
            The chapter's description
        created: str
            The chapter's creation date in datetime form: yyy-mm-dd hh:mm:ss
        modified: str
            The chapter's last modification date in datetime form: yyy-mm-dd hh:mm:ss
        user: User
            The user who owns this entry
        story: Story
            The story that the chapter belongs to
        links: List[ChapterLink]
            The links of the chapter
        notes: List[ChapterNote]
            The notes of the chapter

    Methods
    -------
        __repr__()
            Returns a string representation of the chapter
        __str__()
            Returns a string representation of the chapter
        serialize()
            Returns a dictionary representation of the chapter
        unserialize(data: dict)
            Updates the chapter's attributes with the values from the dictionary
        validate_title(title: str)
            Validates the title's length
        validate_description(description: str)
            Validates the description's length
    """

    __tablename__ = 'chapters'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    story_id: Mapped[int] = mapped_column(Integer, ForeignKey('stories.id'))
    position: Mapped[int] = mapped_column(Integer, nullable=False, default=1)
    title: Mapped[str] = mapped_column(String(250), nullable=False)
    description: Mapped[str] = mapped_column(Text, nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    scenes: Mapped[Optional[List["Scene"]]] = relationship(
        "Scene", back_populates="chapter", lazy="joined",
        cascade="all, delete, delete-orphan")
    user: Mapped["User"] = relationship("User")
    story: Mapped["Story"] = relationship("Story", back_populates="chapters")
    links: Mapped[Optional[List["ChapterLink"]]] = relationship(
        "ChapterLink", back_populates="chapter", lazy="joined",
        cascade="all, delete, delete-orphan")
    notes: Mapped[Optional[List["ChapterNote"]]] = relationship(
        "ChapterNote", back_populates="chapter", lazy="joined",
        cascade="all, delete, delete-orphan")

    def __repr__(self):
        """Returns a string representation of the chapter.

        Returns
        -------
        str
            A string representation of the chapter
        """

        return f'<Chapter {self.position}: {self.title!r}>'

    def __str__(self):
        """Returns a string representation of the chapter.

        Returns
        -------
        str
            A string representation of the chapter
        """

        return f'{self.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the chapter.

        Returns
        -------
        dict
            A dictionary representation of the chapter
        """

        links = []
        if self.links:
            for chapter_link in self.links:
                links.append(chapter_link.link.serialize())

        notes = []
        if self.notes:
            for chapter_note in self.notes:
                notes.append(chapter_note.note.serialize())

        scenes = []
        if self.scenes:
            for scene in self.scenes:
                scenes.append(scene.serialize())

        escaped_title = self.title.replace('"', "'").replace('/', '\\/')
        escaped_description = None

        if self.description:
            escaped_description = self.description.replace('"', "'").replace('/', '\\/')

        return {
            'id': self.id,
            'user_id': self.user_id,
            'story_id': self.story_id,
            'position': self.position,
            'title': escaped_title,
            'description': escaped_description,
            'created': str(self.created),
            'modified': str(self.modified),
            'links': links,
            'notes': notes,
            'scenes': scenes
        }

    def unserialize(self, data: dict) -> "Chapter":
        """Updates the chapter's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the chapter

        Returns
        -------
        Chapter
            The updated chapter
        """

        self.id = data.get('id', self.id)
        self.user_id = data.get('user_id', self.user_id)
        self.story_id = data.get('story_id', self.story_id)
        self.position = data.get('position', self.position)
        self.title = data.get('title', self.title)
        self.description = data.get('description', self.description)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("title")
    def validate_title(self, key, title: str) -> str:
        """Validates the title's length.

        Parameters
        ----------
        title: str
            The chapter's title

        Returns
        -------
        str
            The validated title
        """

        if not title:
            raise ValueError("A chapter title is required.")

        if len(title) > 250:
            raise ValueError("The chapter title must have no more than 250 characters.")

        return title

    @validates("description")
    def validate_description(self, key, description: str) -> str:
        """Validates the description's length.

        Parameters
        ----------
        description: str
            The chapter's description

        Returns
        -------
        str
            The validated description
        """

        if description and len(description) > 65535:
            raise ValueError("The chapter description must have no more than 65535 characters.")

        return description



from enum import Enum


class ImageMimeTypes(Enum):
    """The ImageMimeTypes class represents the types of image mime types.

    Attributes
    ----------
        JPEG: str
            The image is in JPEG format
        PNG: str
            The image is in PNG format
        GIF: str
            The image is in GIF format
    """
    JPEG = 'image/jpeg'
    PNG = 'image/png'
    GIF = 'image/gif'



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from gnatwriter.models import User, Story, Chapter, Link, Base


class ChapterLink(Base):
    """The ChapterLink class represents the relationship between a chapter and a link.

    Attributes
    ----------
        user_id: int
            The id of the owner of this entry
        story_id: int
            The story's id
        chapter_id: int
            The chapter's id
        link_id: int
            The link's id
        created: str
            The creation datetime of the link between the Chapter and the Link
        user: User
            The user who owns this entry
        story: Story
            The story that the chapter belongs to
        chapter: Chapter
            The chapter that the link belongs to
        link: Link
            The link that the chapter has

    Methods
    -------
        __repr__()
            Returns a string representation of the relationship
        __str__()
            Returns a string representation of the relationship
        serialize()
            Returns a dictionary representation of the relationship
        unserialize(data: dict)
            Updates the relationship's attributes with the values from the dictionary
    """

    __tablename__ = 'chapters_links'
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    story_id: Mapped[int] = mapped_column(Integer, ForeignKey('stories.id'))
    chapter_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('chapters.id'), primary_key=True
    )
    link_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('links.id'), primary_key=True
    )
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    user: Mapped["User"] = relationship("User")
    story: Mapped["Story"] = relationship("Story")
    chapter: Mapped["Chapter"] = relationship(
        "Chapter", back_populates="links"
    )
    link: Mapped["Link"] = relationship(
        "Link", back_populates="chapters", lazy="joined"
    )

    def __repr__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'<ChapterLink {self.chapter.title!r} - {self.link.title!r}>'

    def __str__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'{self.chapter.title} - {self.link.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the relationship.

        Returns
        -------
        dict
            A dictionary representation of the relationship
        """

        return {
            'user_id': self.user_id,
            'story_id': self.story_id,
            'chapter_id': self.chapter_id,
            'link_id': self.link_id,
            'created': str(self.created),
        }

    def unserialize(self, data: dict) -> "ChapterLink":
        """Updates the relationship's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the relationship

        Returns
        -------
        ChapterLink
            The updated relationship
        """

        self.user_id = data.get('user_id', self.user_id)
        self.story_id = data.get('story_id', self.story_id)
        self.chapter_id = data.get('chapter_id', self.chapter_id)
        self.link_id = data.get('link_id', self.link_id)
        self.created = data.get('created', self.created)

        return self



from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, Boolean, String, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, AuthorStory, Base


class Author(Base):
    """The Author class represents an author of a story.

    Attributes
    ----------
        id: int
            The author's id
        user_id: int
            The id of the owner of this entry
        is_pseudonym: bool
            Whether the author is a pseudonym or not
        name: str
            The author's name
        initials: str
            The author's initials
        created: str
            The author's creation date in datetime form: yyy-mm-dd hh:mm:ss
        modified: str
            The author's last modification date in datetime form: yyy-mm-dd hh:mm:ss
        user: User
            The user who owns this entry
        stories: List[AuthorStory]
            The stories that the author has written

    Methods
    -------
        __repr__()
            Returns a string representation of the author
        __str__()
            Returns a string representation of the author
        serialize()
            Returns a dictionary representation of the author
        unserialize(data: dict)
            Updates the author's attributes with the values from the dictionary
        validate_name(name: str)
            Validates the name's length
        validate_initials(initials: str)
            Validates the initials' length
    """

    __tablename__ = 'authors'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    is_pseudonym: Mapped[bool] = mapped_column(Boolean, default=False)
    name: Mapped[str] = mapped_column(String(150), nullable=False)
    initials: Mapped[str] = mapped_column(String(10), nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User", back_populates="authors")
    stories: Mapped[Optional[List["AuthorStory"]]] = relationship(
        "AuthorStory", back_populates="author",
        cascade="all, delete, delete-orphan", lazy="joined")

    def __repr__(self):
        """Returns a string representation of the author.

        Returns
        -------
        str
            A string representation of the author
        """

        return f'<Author {self.name!r}>'

    def __str__(self):
        """Returns a string representation of the author.

        Returns
        -------
        str
            A string representation of the author
        """

        return f'Author: {self.name}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the author.

        Returns
        -------
        dict
            A dictionary representation of the author
        """

        return {
            'id': self.id,
            'user_id': self.user_id,
            'is_pseudonym': self.is_pseudonym,
            'name': self.name,
            'initials': self.initials,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "Author":
        """Updates the author's attributes with the values from the dictionary.

        Returns
        -------
        Author
            The updated author
        """

        self.id = data.get('id', self.id)
        self.user_id = data.get('user_id', self.user_id)
        self.is_pseudonym = data.get('is_pseudonym', self.is_pseudonym)
        self.name = data.get('name', self.name)
        self.initials = data.get('initials', self.initials)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("name")
    def validate_name(self, key, name: str) -> str:
        """Validates the name's length.

        Returns
        -------
        str
            The validated name
        """

        if not name:
            raise ValueError("The author name must not be empty.")

        if len(name) > 150:
            raise ValueError("The author's name must not have more than 150 characters.")

        return name

    @validates("initials")
    def validate_initials(self, key, initials: str) -> str:
        """Validates the initials' length.

        Returns
        -------
        str
            The validated initials
        """

        if initials and len(initials) > 10:
            raise ValueError("The author's initials must not have more than 10 characters.")

        return initials



from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, String, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, LinkStory, ChapterLink, LinkScene, CharacterLink, EventLink, LinkLocation, Base
from validators import url as url_validator


class Link(Base):
    """The Link class represents a web link in the application.

    Attributes
    ----------
        id: int
            The link's id
        user_id: int
            The id of the owner of this entry
        title: str
            The link's title
        url: str
            The link's URL
        created: str
            The creation datetime of the link
        modified: str
            The last modification datetime of the link
        user: User
            The user who owns this entry
        stories: List[LinkStory]
            The stories that the link has
        chapters: List[ChapterLink]
            The chapters that the link has
        scenes: List[LinkScene]
            The scenes that the link has
        characters: List[CharacterLink]
            The characters that the link has
        events: List[EventLink]
            The events that the link has
        locations: List[LinkLocation]
            The locations that the link has

    Methods
    -------
        __repr__()
            Returns a string representation of the link
        __str__()
            Returns a string representation of the link
        serialize()
            Returns a dictionary representation of the link
        unserialize(data: dict)
            Updates the link's attributes with the values from the dictionary
        validate_title(title: str)
            Validates the title's length
        validate_url(url: str)
            Validates the URL's length and format
    """

    __tablename__ = 'links'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    title: Mapped[str] = mapped_column(String(250), nullable=False)
    url: Mapped[str] = mapped_column(String(200), nullable=False)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User", back_populates="links")
    stories: Mapped[Optional[List["LinkStory"]]] = relationship(
        "LinkStory", back_populates="link",
        cascade="all, delete, delete-orphan")
    chapters: Mapped[Optional[List["ChapterLink"]]] = relationship(
        "ChapterLink", back_populates="link",
        cascade="all, delete, delete-orphan")
    scenes: Mapped[Optional[List["LinkScene"]]] = relationship(
        "LinkScene", back_populates="link",
        cascade="all, delete, delete-orphan")
    characters: Mapped[Optional[List["CharacterLink"]]] = relationship(
        "CharacterLink", back_populates="link",
        cascade="all, delete, delete-orphan")
    events: Mapped[Optional[List["EventLink"]]] = relationship(
        "EventLink", back_populates="link",
        cascade="all, delete, delete-orphan")
    locations: Mapped[Optional[List["LinkLocation"]]] = relationship(
        "LinkLocation", back_populates="link",
        cascade="all, delete, delete-orphan")

    def __repr__(self):
        """Returns a string representation of the link.

        Returns
        -------
        str
            A string representation of the link
        """

        return f'<Link {self.title!r}>'

    def __str__(self):
        """Returns a string representation of the link.

        Returns
        -------
        str
            A string representation of the link
        """

        return f'{self.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the link.

        Returns
        -------
        dict
            A dictionary representation of the link
        """

        escaped_title = self.title.replace('"', '\\"').replace('/', '\\/')
        escaped_url = self.description.replace('"', '\\"').replace('/', '\\/')

        return {
            'id': self.id,
            'user_id': self.user_id,
            'title': escaped_title,
            'url': escaped_url,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "Link":
        """Updates the link's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the link

        Returns
        -------
        Link
            The unserialized link
        """

        self.user_id = data.get('user_id', self.user_id)
        self.title = data.get('title', self.title)
        self.url = data.get('url', self.url)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("title")
    def validate_title(self, key, title: str) -> str:
        """Validates the title's length.

        Parameters
        ----------
        title: str
            The link's title

        Returns
        -------
        str
            The validated title
        """

        if not title:
            raise ValueError("A link title is required.")

        if len(title) > 250:
            raise ValueError("The link title can have no more than 250 characters.")

        return title

    @validates("url")
    def validate_url(self, key, url: str) -> str:
        """Validates the URL's length and format.

        Parameters
        ----------
        url: str
            The link's URL

        Returns
        -------
        str
            The validated URL
        """

        if not url:
            raise ValueError("A link URL is required.")

        if len(url) > 200:
            raise ValueError("The link URL can have no more than 200 characters.")

        if not url_validator(url):
            raise ValueError("The link URL is not valid.")

        return url



from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.controllers.ActivityController import ActivityController
from gnatwriter.controllers.AssistantController import AssistantController
from gnatwriter.controllers.AuthorController import AuthorController
from gnatwriter.controllers.BibliographyController import BibliographyController
from gnatwriter.controllers.ChapterController import ChapterController
from gnatwriter.controllers.CharacterController import CharacterController
from gnatwriter.controllers.EventController import EventController
from gnatwriter.controllers.ExportController import ExportController
from gnatwriter.controllers.ImageController import ImageController
from gnatwriter.controllers.LinkController import LinkController
from gnatwriter.controllers.LocationController import LocationController
from gnatwriter.controllers.NoteController import NoteController
from gnatwriter.controllers.OllamaModelController import OllamaModelController
from gnatwriter.controllers.SceneController import SceneController
from gnatwriter.controllers.StoryController import StoryController
from gnatwriter.controllers.SubmissionController import SubmissionController
from gnatwriter.controllers.UserController import UserController








from datetime import datetime
from sqlalchemy import Integer, String, DateTime, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, Base


class Activity(Base):
    """The Activity class documents any controller endpoint activities.

    Attributes
    ----------
        id: int
            The activity's id
        user_id: int
            The id of the owner of this entry
        summary: str
            The activity's summary
        created: str
            The activity's creation date in datetime form: yyy-mm-dd hh:mm:ss
        user: User
            The system user or user currently logged in

    Methods
    -------
        __repr__()
            Returns a string representation of the activity
        __str__()
            Returns a string representation of the activity
        serialize()
            Returns a dictionary representation of the activity
        unserialize(data: dict)
            Updates the activity's attributes with the values from the dictionary
        validate_summary(summary: str)
            Validates the summary's length
    """

    __tablename__ = 'activities'
    id: Mapped[int] = mapped_column(
        Integer, primary_key=True, autoincrement=True
    )
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    summary: Mapped[str] = mapped_column(String(250), nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    user: Mapped["User"] = relationship(
        "User", back_populates="activities"
    )

    def __repr__(self):
        """Returns a string representation of the activity.

        Returns
        -------
        str
            A string representation of the activity
        """

        return f'<Activity {self.summary!r}>'

    def __str__(self):
        """Returns a string representation of the activity.

        Returns
        -------
        str
            A string representation of the activity
        """

        return f'{self.summary}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the activity.

        Returns
        -------
        dict
            A dictionary representation of the activity
        """

        return {
            'id': self.id,
            'user_id': self.user_id,
            'summary': self.summary,
            'created': str(self.created),
        }

    def unserialize(self, data: dict) -> "Activity":
        """Updates the activity's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the activity

        Returns
        -------
        Activity
            The updated activity
        """

        self.id = data.get('id', self.id)
        self.user_id = data.get('user_id', self.user_id)
        self.summary = data.get('summary', self.summary)
        self.created = data.get('created', self.created)

        return self

    @validates("summary")
    def validate_summary(self, key, summary: str) -> str:
        """Validates the summary's length.
        """

        if len(summary) > 250:
            raise ValueError(
                """The activity summary must be no more than 250 characters in length.
                """
            )

        return summary



from configparser import ConfigParser
from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, String, Date, Text, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, CharacterRelationship, CharacterEvent, CharacterTrait, CharacterImage, CharacterLink, \
    CharacterNote, CharacterStory, Base

mbti_types = [
    "INTJ", "INTP", "ENTJ", "ENTP",
    "INFJ", "INFP", "ENFJ", "ENFP",
    "ISTJ", "ISFJ", "ESTJ", "ESFJ",
    "ISTP", "ISFP", "ESTP", "ESFP"
]
enneagram_types = [
    "Type 1", "Type 2", "Type 3", "Type 4",
    "Type 5", "Type 6", "Type 7", "Type 8", "Type 9"
]


class Character(Base):
    """The Character class represents a character in one or more stories.

    Attributes
    ----------
        id: int
            The Character ID
        user_id: int
            The id of the owner of this entry
        title: str
            The character's title
        honorific: str
            The character's honorific
        first_name: str
            The character's first name
        middle_name: str
            The character's middle name
        last_name: str
            The character's last name
        nickname: str
            The character
        gender: str
            The gender of the character
        sex: str
            The sex of the character
        ethnicity: str
            The ethnicity of the character
        race: str
            The race of the character
        tribe_or_clan: str
            The tribe or clan of the character
        nationality: str
            The nationality of the character
        religion: str
            The religion of the character
        occupation: str
            The occupation of the character
        education: str
            The education of the character
        marital_status: str
            The marital status of the character
        children: bool
            Whether the character has children
        date_of_birth: str
            The character's date of birth in date form: yyyy-mm-dd
        date_of_death: str
            The character's date of death in date form: yyyy-mm-dd
        description: str
            The character's description
        mbti: str
            The Myers-Briggs Type Indicator of the character
        enneagram: str
            The Enneagram type of the character
        wounds: str
            The character's wounds
        created: str
            The character's creation date in datetime form: yyy-mm-dd hh:mm:ss
        modified: str
            The character's last modification date in datetime form: yyy-mm-dd hh:mm:ss

    Methods
    -------
        __repr__()
            Returns a string representation of the character
        __str__()
            Returns a string representation of the character
        serialize()
            Returns a dictionary representation of the character
        unserialize(data: dict)
            Updates the character's attributes with the values from the dictionary
        validate_title(title: str)
            Validates the title's length
        validate_honorific(honorific: str)
            Validates the honorific's length
        validate_first_name(first_name: str):
            Validates the first name's length
        validate_middle_name(middle_name: str)
            Validates the middle name's length
        validate_last_name(last_name: str)
            Validates the last name's length
        validate_nickname(nickname: str)
            Validates the nickname's length
        validate_gender(gender: str)
            Validates the gender's length
        validate_sex(sex: str)
            Validates the sex's length
        validate_ethnicity(ethnicity: str)
            Validates the ethnicity's length
        validate_race(race: str)
            Validates the length of the race of the character
        validate_tribe_or_clan(tribe_or_clan: str)
            Validates the length of tribe or clan of the character
        validate_nationality(nationality: str)
            Validates the length of the nationality value
        validate_religion(religion: str)
            Validates the length of the religion value
        validate_occupation(occupation: str)
            Validates the length of the occupation value
        validate_education(education: str)
            Validates the length of the education value
        validate_marital_status(marital_status: str)
            Validates the length of the marital status value
        validate_date_of_birth(date_of_birth: str)
            Validates the date of birth's format
        validate_date_of_death(date_of_death: str)
            Validates the date of death's format
        validate_description(description: str)
            Validates the description's length
        validate_mbti(mbti: str)
            Validates the MBTI type of the character
        validate_enneagram(enneagram: str)
            Validates the Enneagram type of the character
        validate_wounds(wounds: str)
            Validates the wounds' length
    """

    __tablename__ = 'characters'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    title: Mapped[str] = mapped_column(String(100), nullable=True)
    honorific: Mapped[str] = mapped_column(String(50), nullable=True)
    first_name: Mapped[str] = mapped_column(String(100), nullable=True)
    middle_name: Mapped[str] = mapped_column(String(100), nullable=True)
    last_name: Mapped[str] = mapped_column(String(100), nullable=True)
    nickname: Mapped[str] = mapped_column(String(100), nullable=True)
    gender: Mapped[str] = mapped_column(String(50), nullable=True)
    sex: Mapped[str] = mapped_column(String(50), nullable=True)
    ethnicity: Mapped[str] = mapped_column(String(50), nullable=True)
    race: Mapped[str] = mapped_column(String(50), nullable=True)
    tribe_or_clan: Mapped[str] = mapped_column(String(50), nullable=True)
    nationality: Mapped[str] = mapped_column(String(50), nullable=True)
    religion: Mapped[str] = mapped_column(String(50), nullable=True)
    occupation: Mapped[str] = mapped_column(String(50), nullable=True)
    education: Mapped[str] = mapped_column(Text, nullable=True)
    marital_status: Mapped[str] = mapped_column(String(50), nullable=True)
    children: Mapped[bool] = mapped_column(Integer, nullable=True)
    date_of_birth: Mapped[str] = mapped_column(Date, nullable=True)
    date_of_death: Mapped[str] = mapped_column(Date, nullable=True)
    description: Mapped[str] = mapped_column(Text, nullable=True)
    mbti: Mapped[str] = mapped_column(String(50), nullable=True)
    enneagram: Mapped[str] = mapped_column(String(50), nullable=True)
    wounds: Mapped[str] = mapped_column(Text, nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User", back_populates="characters")
    character_relationships: Mapped[Optional[List["CharacterRelationship"]]] = relationship(
        "CharacterRelationship", back_populates="related_character",
        foreign_keys="[CharacterRelationship.related_id]", lazy="joined",
        cascade="all, delete, delete-orphan"
    )
    traits: Mapped[Optional[List["CharacterTrait"]]] = relationship(
        "CharacterTrait", back_populates="character", lazy="joined",
        cascade="all, delete, delete-orphan"
    )
    events: Mapped[Optional[List["CharacterEvent"]]] = relationship(
        "CharacterEvent", back_populates="character",
        cascade="all, delete, delete-orphan", lazy="joined"
    )
    images: Mapped[Optional[List["CharacterImage"]]] = relationship(
        "CharacterImage", back_populates="character", lazy="joined",
        cascade="all, delete, delete-orphan"
    )
    links: Mapped[Optional[List["CharacterLink"]]] = relationship(
        "CharacterLink", back_populates="character", lazy="joined",
        cascade="all, delete, delete-orphan"
    )
    notes: Mapped[Optional[List["CharacterNote"]]] = relationship(
        "CharacterNote", back_populates="character", lazy="joined",
        cascade="all, delete, delete-orphan"
    )
    stories: Mapped[Optional[List["CharacterStory"]]] = relationship(
        "CharacterStory", back_populates="character", lazy="joined",
        cascade="all, delete, delete-orphan"
    )

    def __repr__(self):
        """Returns a string representation of the character.

        Returns
        -------
        str
            A string representation of the character
        """

        return f'<Character {self.title!r} {self.first_name!r} {self.last_name!r}>'

    def __str__(self):
        """Returns a string representation of the character.

        Returns
        -------
        str
            A string representation of the character
        """

        title = f'{self.title}' if self.title else ""
        first_name = f' {self.first_name}' if self.first_name else ""
        middle_name = f' {self.middle_name}' if self.middle_name else ""
        last_name = f' {self.last_name}' if self.last_name else ""
        nickname = f' ({self.nickname})' if self.nickname else ""

        return f'{title}{first_name}{middle_name}{last_name}{nickname}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the character.

        Returns
        -------
        dict
            A dictionary representation of the character
        """

        relationships = []
        for character_relationship in self.character_relationships:
            relationships.append(
                character_relationship.serialize()
            )

        traits = []
        for trait in self.traits:
            traits.append(
                trait.serialize()
            )

        images = []
        for character_image in self.images:
            images.append(
                character_image.serialize()
            )

        events = []
        for event in self.events:
            events.append(
                event.serialize()
            )

        links = []
        for character_link in self.links:
            links.append(
                character_link.serialize()
            )

        notes = []
        for character_note in self.notes:
            notes.append(
                character_note.serialize()
            )

        stories = []
        for character_story in self.stories:
            stories.append(
                character_story.serialize()
            )

        return {
            'id': self.id,
            'user_id': self.user_id,
            'title': self.title,
            'honorific': self.honorific,
            'first_name': self.first_name,
            'middle_name': self.middle_name,
            'last_name': self.last_name,
            'nickname': self.nickname,
            'full_name': self.full_name,
            'gender': self.gender,
            'sex': self.sex,
            'ethnicity': self.ethnicity,
            'race': self.race,
            'tribe_or_clan': self.tribe_or_clan,
            'nationality': self.nationality,
            'religion': self.religion,
            'occupation': self.occupation,
            'education': self.education,
            'marital_status': self.marital_status,
            'children': self.children,
            'age': self.age,
            'date_of_birth': str(self.date_of_birth),
            'date_of_death': str(self.date_of_death),
            'description' : self.description,
            'mbti': self.mbti,
            'enneagram': self.enneagram,
            'wounds': self.wounds,
            'created': str(self.created),
            'modified': str(self.modified),
            'relationships': relationships,
            'traits': traits,
            'events': events,
            'images': images,
            'links': links,
            'notes': notes,
            'stories': stories
        }

    def unserialize(self, data: dict) -> "Character":
        """Updates the character's attributes with the values from the dictionary.

        Returns
        -------
        Character
            The updated character
        """

        self.id = data.get('id', self.id)
        self.user_id = data.get('user_id', self.user_id)
        self.title = data.get('title', self.title)
        self.honorific = data.get('honorific', self.honorific)
        self.first_name = data.get('first_name', self.first_name)
        self.middle_name = data.get('middle_name', self.middle_name)
        self.last_name = data.get('last_name', self.last_name)
        self.nickname = data.get('nickname', self.nickname)
        self.gender = data.get('gender', self.gender)
        self.sex = data.get('sex', self.sex)
        self.ethnicity = data.get('ethnicity', self.ethnicity)
        self.nationality = data.get('nationality', self.nationality)
        self.religion = data.get('religion', self.religion)
        self.occupation = data.get('occupation', self.occupation)
        self.education = data.get('education', self.education)
        self.marital_status = data.get('marital_status', self.marital_status)
        self.children = data.get('children', self.children)
        self.date_of_birth = data.get('date_of_birth', self.date_of_birth)
        self.date_of_death = data.get('date_of_death', self.date_of_death)
        self.description = data.get('description', self.description)
        self.mbti = data.get('mbti', self.mbti)
        self.enneagram = data.get('enneagram', self.enneagram)
        self.wounds = data.get('wounds', self.wounds)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("title")
    def validate_title(self, key, title: str) -> str:
        """Validates the title's length.

        Parameters
        ----------
        title: str
            The character's title

        Returns
        -------
        str
            The validated title
        """

        if title and len(title) > 100:
            raise ValueError("The character title must have no more than 100 characters.")

        return title

    @validates("honorific")
    def validate_honorific(self, key, honorific: str) -> str:
        """Validates the honorific's length.

        Parameters
        ----------
        honorific: str
            The character's honorific

        Returns
        -------
        str
            The validated honorific
        """

        if honorific and len(honorific) > 50:
            raise ValueError("The value of honorific must have no more than 50 characters.")

        return honorific

    @validates("first_name")
    def validate_first_name(self, key, first_name: str) -> str:
        """Validates the first name's length.

        Parameters
        ----------
        first_name: str
            The character's first name

        Returns
        -------
        str
            The validated first name
        """

        if first_name and len(first_name) > 100:
            raise ValueError("The character first name must have no more than 100 characters.")

        return first_name

    @validates("middle_name")
    def validate_middle_name(self, key, middle_name: str) -> str:
        """Validates the middle name's length.

        Parameters
        ----------
        middle_name: str
            The character's middle name

        Returns
        -------
        str
            The validated middle name
        """

        if middle_name and len(middle_name) > 100:
            raise ValueError("The character middle name must have no more than 100 characters.")

        return middle_name

    @validates("last_name")
    def validate_last_name(self, key, last_name: str) -> str:
        """Validates the last name's length.

        Parameters
        ----------
        last_name: str
            The character's last name

        Returns
        -------
        str
            The validated last name
        """

        if last_name and len(last_name) > 100:
            raise ValueError("The character last name must have no more than 100 characters.")

        return last_name

    @validates("nickname")
    def validate_nickname(self, key, nickname: str) -> str:
        """Validates the nickname's length.

        Parameters
        ----------
        nickname: str
            The character's nickname

        Returns
        -------
        str
            The validated nickname
        """

        if nickname and len(nickname) > 100:
            raise ValueError("The character nickname must have no more than 100 characters.")

        return nickname

    @validates("gender")
    def validate_gender(self, key, gender: str) -> str:
        """Validates the gender's length.

        Parameters
        ----------
        gender: str
            The character's gender

        Returns
        -------
        str
            The validated gender
        """

        if gender and len(gender) > 50:
            raise ValueError("The gender value must have no more than 50 characters.")

        return gender

    @validates("sex")
    def validate_sex(self, key, sex: str) -> str:
        """Validates the sex's length.

        Parameters
        ----------
        sex: str
            The character's sex

        Returns
        -------
        str
            The validated sex
        """

        if sex and len(sex) > 50:
            raise ValueError("The value of sex must have no more than 50 characters.")

        return sex

    @validates("ethnicity")
    def validate_ethnicity(self, key, ethnicity: str) -> str:
        """Validates the ethnicity's length.

        Parameters
        ----------
        ethnicity: str
            The character's ethnicity

        Returns
        -------
        str
            The validated ethnicity
        """

        if ethnicity and len(ethnicity) > 50:
            raise ValueError("The value of ethnicity must have no more than 50 characters.")

        return ethnicity

    @validates("race")
    def validate_race(self, key, race: str) -> str:
        """Validates the length of race of the character.

        Parameters
        ----------
        race: str
            The race of the character

        Returns
        -------
        str
            The validated race
        """

        if race and len(race) > 50:
            raise ValueError("The value of race must have no more than 50 characters.")

        return race

    @validates("tribe_or_clan")
    def validate_tribe_or_clan(self, key, tribe_or_clan: str) -> str:
        """Validates the length of tribe or clan of the character.

        Parameters
        ----------
        tribe_or_clan: str
            The tribe or clan of the character

        Returns
        -------
        str
            The validated tribe or clan
        """

        if tribe_or_clan and len(tribe_or_clan) > 50:
            raise ValueError("The value of tribe or clan must have no more than 50 characters.")

        return tribe_or_clan

    @validates("nationality")
    def validate_nationality(self, key, nationality: str) -> str:
        """Validates the length of the nationality value.

        Parameters
        ----------
        nationality: str
            The nationality of the character

        Returns
        -------
        str
            The character's validated nationality
        """

        if nationality and len(nationality) > 50:
            raise ValueError("The value of nationality must have no more than 50 characters.")

        return nationality

    @validates("religion")
    def validate_religion(self, key, religion: str) -> str:
        """Validates the length of the religion value.

        Parameters
        ----------
        religion: str
            The religion of the character

        Returns
        -------
        str
            The character's validated religion
        """

        if religion and len(religion) > 50:
            raise ValueError("The value of religion must have no more than 50 characters.")

        return religion

    @validates("occupation")
    def validate_occupation(self, key, occupation: str) -> str:
        """Validates the length of the occupation value.

        Parameters
        ----------
        occupation: str
            The occupation of the character

        Returns
        -------
        str
            The character's validated occupation
        """

        if occupation and len(occupation) > 50:
            raise ValueError("The value of occupation must have no more than 50 characters.")

        return occupation

    @validates("education")
    def validate_education(self, key, education: str) -> str:
        """Validates the length of the education value.

        Parameters
        ----------
        education: str
            The education of the character

        Returns
        -------
        str
            The character's validated education
        """

        if education and len(education) > 65535:
            raise ValueError("The value of education must have no more than 65535 characters.")

        return education

    @validates("marital_status")
    def validate_marital_status(self, key, marital_status: str) -> str:
        """Validates the length of the marital status value.

        Parameters
        ----------
        marital_status: str
            The marital status of the character

        Returns
        -------
        str
            The character's validated marital status
        """

        if marital_status and len(marital_status) > 50:
            raise ValueError("The value of marital status must have no more than 50 characters.")

        return marital_status

    @validates("description")
    def validate_description(self, key, description: str) -> str:
        """Validates the description's length.

        Parameters
        ----------
        description: str
            The character's description

        Returns
        -------
        str
            The validated description
        """

        if description and len(description) > 65535:
            raise ValueError("The character description must have no more than 65535 characters.")

        return description

    @validates("mbti")
    def validate_mbti(self, key, mbti: str) -> str:
        """Validates the MBTI type of the character.

        Parameters
        ----------
        mbti: str
            The MBTI type of the character

        Returns
        -------
        str
            The validated MBTI type
        """

        if mbti and mbti not in mbti_types:
            raise ValueError("The MBTI type must be one of the following: INTJ, INTP, ENTJ, ENTP, INFJ, INFP, ENFJ, ENFP, ISTJ, ISFJ, ESTJ, ESFJ, ISTP, ISFP, ESTP, ESFP")

        return mbti

    @validates("enneagram")
    def validate_enneagram(self, key, enneagram: str) -> str:
        """Validates the Enneagram type of the character.

        Parameters
        ----------
        enneagram: str
            The Enneagram type of the character

        Returns
        -------
        str
            The validated Enneagram type
        """

        if enneagram and enneagram not in enneagram_types:
            raise ValueError("The Enneagram type must be one of the following: Type 1, Type 2, Type 3, Type 4, Type 5, Type 6, Type 7, Type 8, Type 9")

        return enneagram

    @validates("wounds")
    def validate_wounds(self, key, wounds: str) -> str:
        """Validates the wounds' length.

        Parameters
        ----------
        wounds: str
            The character's wounds

        Returns
        -------
        str
            The validated wounds
        """

        if wounds and len(wounds) > 65535:
            raise ValueError("The character wounds must have no more than 65535 characters.")

        return wounds

    @property
    def age(self) -> int:
        """Returns the character's age.

        Returns
        -------
        int
            The character's age
        """

        if self.date_of_birth:
            if self.date_of_death:
                return self.date_of_death.year - self.date_of_birth.year

            else:
                return datetime.now().year - self.date_of_birth.year

        else:
            return 0

    @property
    def full_name(self) -> str:
        """Returns the character's full name.

        Returns
        -------
        str
            The character's full name
        """

        full_name = ""

        if self.title:
            full_name += f'({self.title})'

        if self.honorific:
            if len(full_name) > 0:
                full_name += " "
            full_name += f'{self.honorific}'

        if self.first_name:
            if len(full_name) > 0:
                full_name += " "
            full_name += f'{self.first_name}'

        if self.middle_name:
            if self.first_name or self.title:
                full_name += " "
            full_name += f'{self.middle_name}'

        if self.last_name:
            if self.middle_name or self.first_name or self.title:
                full_name += " "
            full_name += f'{self.last_name}'

        if self.nickname:
            if len(full_name) > 0:
                full_name += " "
            full_name += f'({self.nickname})'

        return f'{full_name}'
